/* Uproszczone operacje semaforowe */

#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/sem.h>

#include <errno.h>

extern int errno;

#define BIGCOUNT 10000 // wartosc poczatkowa licznika procesow

/* definicjia tablic dotyczacych operacji na semaforach */
static struct sembuf op_lock[2] = {
    2, 0, 0, //czekaj az [2] (zajetosc) rowna sie 0
    2, 1, SEM_UNDO  //po czym przypisz wartosc 1 (zajmujac go)
		    //SEM_UNDO by zwolnic semafor jak proces padnie
};

static struct sembuf op_endcreate[2] = {
    1, -1, SEM_UNDO, //zmniejsz semafor [1] (licznik procesow)
    2, -1, SEM_UNDO //nastepnie wyzeruj semafor [2] (zajetosc)
};

static struct sembuf op_open[1] = {
    1, -1, SEM_UNDO //zmniejsz semafor [1] (licznik procesow)
};

static struct sembuf op_close[3] = {
    2, 0, 0, // czekaj az [2] (zajetosc rowna sie 0
    2, 1, SEM_UNDO, // po czym przypisz mu wartosc 1 (zajmujac go)
    1, 1, SEM_UNDO // /nasepnie zwiesz semafor [1] (licznik proc.)
};

static struct sembuf op_unlock[1] = {
    2, -1, SEM_UNDO //wyzeruj semafor [2] (zajetosc)
};

static struct sembuf op_op[1] = {
    0, 99, SEM_UNDO // zmniejsz lub zwieksz semafor [0]
		    // w miejsce wartosci 99 bedzie przypisana 
		    // rzeczywista liczba dodawana lub odejmowana
		    // (dodatnia lub ujemna)
};

/* Tworz semafor przypisujac mu okreslona wartosc poczatkowa */
/* przekaz identyfikator semafora, jak blad to -1 */
int sem_create(key, initval)
key_t key;
int initval; 
{
register int id,semval;
union semun {
    int val;
    struct semid_ds *buf;
    ushort *array;
} semctl_arg;

if (key == IPC_PRIVATE)
    return(-1); //nie udostepniamy semaforow prywatnych
if (key == (key_t) -1)
    return(-1); 
again:
    if ((id = semget(key, 3, 0666 |IPC_CREAT)) <0)
	return(-1); //klopoty z uprawnieniami lub przepelnienie tablicy
/* jezeli jakis proces w miedzy czasie usunie semafor przez sem_close() to trzeba to obsluzyc */
if (semop(id, op_lock, 2) <0 )
    {
    if (errno == EINVAL)
	goto again;
    perror("nie moge zalokowac");
    }
/*  */
/* pobierz wartosc licznika procesow. Jezeli jest zerem, to nikt jeszcze nie zainicjiowal semafora */
if ((semval = semctl(id, 1, GETVAL, 0)) <0)
    perror("nie moze pobrac wartosci licznika procesow");
if (semval == 0)
    {
/*     opcja SETALL mogla by wyzerowac wartosc nastanwa semafora */
    semctl_arg.val = initval;
    if (semctl(id, 0, SETVAL, semctl_arg) <0)
	perror("nie moge ustawic wartosci [0]");
    semctl_arg.val = BIGCOUNT;
    if (semctl(id, 1, SETVAL, semctl_arg) <0)
	perror("nie moge ustawic wartosci [1]");
    }
/* zmniejsz licznik procesow po czym uwolnij semafor     */
if (semop(id, op_endcreate, 2) <0)
    perror("niemoge stworzycd");
return(id);
}

/* otworz semafor, ktory musi juz istniec */
/* dostajemy id semafora lub -1 jak blad */
int sem_open(key_t key)
{register int id;
if (key == IPC_PRIVATE)
    return(-1); // nie udostepniamy semaforow prywatnych
else if (key == (key_t) -1)
	return(-1); // prawdopodobnie blad spowodowany wywolaniem
		    // ftok()  przez proces wywolujacy
if ((id = semget(key, 3, 0)) < 0)
    return(-1); // nie istnieje lub przepelnienie tablicy
/* zmiejsz licznik procesow */
if (semop(id, op_open, 1) <0)
    perror("nie moge otworzyc semafora");
return(id);
}

/* usun semafor */
sem_rm(id)
int id;
{
if (semctl(id, 0, IPC_RMID, 0) < 0)
    perror("nie moge usunac semafora");
}

/* zamknij semafor */
void sem_close(int id)
{
register int semval;
if (semop(id, op_close, 3) <0)
    perror("nie moge wykonac operacji na sem");
if ((semval = semctl(id, 1, GETVAL, 0)) <0)
    perror("nie moge pobrac wartosci");
if (semval > BIGCOUNT)
    perror("sem[1] > BIGCOUNT");
else 
    if (semval == BIGCOUNT)
	sem_rm(id);
    else
	if (semop(id, op_unlock, 1) <0)
	    perror ("nie moge unlokowac");

}

/* czekaj az wartosc semafora bedzie wieksza niz zero, po czym zmniejsz  */
/* ja o 1 i przekaz sterowanie.  */
/* Operacja P Dikjkstry (opuszczenie semafora) */
sem_wait(id)
int id;
{
sem_op(id,-1);
}

/* zwieksz wartosc semafora o 1.  */
/* Operacja V Dijkstry. (podniesienie semafora) */
sem_signal(id)
int id;
{
sem_op(id, 1);
}

/* ogolna poperacja semaforowa. Zwieksz lub zmniejsz o liczbe */
/* okreslona przez uzytkownika (dodatnia lub ujemna, ale nie zerowa */
sem_op(id, value)
int id;
int value;
{
if ((op_op[0].sem_op = value) == 0)
    perror("Nie moze miec wartosci == 0 ");
if (semop(id, op_op, 1) < 0)
    perror("blad sem_op");
}

/* operacja nie semaforowa z zalozenia nie poprawna ale do testow dobra */
int sem_get_val(id)
{
int ret=-1;
if ((ret = semctl(id, 0, GETVAL, 0)) <0)
    perror("nie moze pobrac wartosci semafora");
return ret;
}

